#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <netinet/in.h>
#include <gtk/gtk.h>
#include <jmp.h>
#include <heap_dump.h>
#include <obj.h>
#include <cls.h>
#include <hash.h>
#include <ui.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <jmp.h>

/* id of last heap dump file. */
static int heap_dump_id = -1;
static hashtab* last_heap = NULL;
static char buf[256];

static int heap_hash_cmp (void* ol1, void* ol2) {
    object_link* o1 = (object_link*)ol1;
    object_link* o2 = (object_link*)ol2;
    if (o1->obj == o2->obj)
	return 0;
    return (int)(o1->obj) - (int)(o2->obj);
}

static size_t heap_hash_f (void* obj, size_t size) {
    object_link* ol = (object_link*)obj;
    return ((size_t)ol->obj) % size;
}

static FILE* get_heap_dump_file () {
    FILE* fp;
    snprintf (buf, 256, "jmp_heap_dump_%d", ++heap_dump_id);
    fp = fopen (buf, "w");
    snprintf (buf, 256, _("createing heap dump file: jmp_heap_dump_%d"), heap_dump_id);
    set_status (buf);
    return fp;
}

static void clear_and_set_last_heap (hashtab* heap_hash) {
    if (last_heap) {
	hash_for_each (free, last_heap);
	hash_free (last_heap);
    }
    last_heap = heap_hash;
}

static void link_object (jobjectID obj, jobjectID parent, jobjectID clz, 
			 int type, jint variable, jint index, 
			 hashtab* heap_hash) {
    object_link* ol = malloc (sizeof (object_link));
    ol->obj = obj;
    ol->parent = parent;
    ol->clz = clz;
    ol->next = hash_search (ol, heap_hash);
    ol->type = type;
    ol->variable = variable;
    ol->index = index;
    hash_insert (ol, heap_hash);    
}

static void* get_pointer (char** p) {
    jobjectID oid;
    memcpy (&oid, *p, sizeof (void*));
    *p += sizeof (void*);    
    return oid;
}

static jint get_u4 (char** p) {
    jint i;
    memcpy (&i, *p, sizeof (i));
    *p += sizeof (i);
    return ntohl (i);
}

static jint get_u2 (char** p) {
    unsigned short u2;
    memcpy (&u2, *p, sizeof (u2));
    *p += sizeof (u2);
    return ntohs (u2);
}

static unsigned char get_u1 (char** p) {
    unsigned char u1;
    memcpy (&u1, *p, sizeof (u1));
    *p += sizeof (u1);
    return u1;
}

static void dump_triplet (FILE* fp, char* prefix, char** p) {
    jobjectID oid = get_pointer (p);
    JNIEnv* env = get_pointer (p);
    jint u4 = get_u4 (p);
    fprintf (fp, "%s: oid=%p, env=%p, frame#=%d\n", prefix, oid, env, u4);
}

static void dump_booleans (FILE* fp, jobjectID oid, jint size, char** p) {
    int i;
    fprintf (fp, "oid=%p[%d] => Z{", oid, size);
    for (i = 0; i < size; i++) {  
	unsigned char u1 = get_u1 (p);
	if (i > 0)
	    fprintf (fp, ", ");
	fprintf (fp, "%d", u1);
    }
    fprintf (fp, "}\n");
}

static void dump_bytes (FILE* fp, jobjectID oid, jint size, char** p) {
    int i;
    fprintf (fp, "oid=%p[%d] => B{", oid, size);
    for (i = 0; i < size; i++) {  
	unsigned char u1 = get_u1 (p);
	if (i > 0)
	    fprintf (fp, ", ");
	fprintf (fp, "%d", u1);
    }
    fprintf (fp, "}\n");
}

static void dump_chars (FILE* fp, jobjectID oid, jint size, char** p) {
    int i;
    fprintf (fp, "oid=%p[%d] => C{", oid, size);
    for (i = 0; i < size; i++) {  
	unsigned short u2 = get_u2 (p);
	if (i > 0)
	    fprintf (fp, ", ");
	fprintf (fp, "%d", u2);
    }
    fprintf (fp, "}\n");    
}

static void dump_shorts (FILE* fp, jobjectID oid, jint size, char** p) {
    int i;
    fprintf (fp, "oid=%p[%d] => S{", oid, size);
    for (i = 0; i < size; i++) {  
	unsigned short u2 = get_u2 (p);
	if (i > 0)
	    fprintf (fp, ", ");
	fprintf (fp, "%d", u2);
    }
    fprintf (fp, "}\n"); 
}

static void dump_ints (FILE* fp, jobjectID oid, jint size, char** p) {
    int i;
    fprintf (fp, "oid=%p[%d] => I{", oid, size);
    for (i = 0; i < size; i++) {  
	jint u4 = get_u4 (p);
	if (i > 0)
	    fprintf (fp, ", ");
	fprintf (fp, "%d", u4);
    }
    fprintf (fp, "}\n"); 
}

static void dump_longs (FILE* fp, jobjectID oid, jint size, char** p) {
    int i;
    fprintf (fp, "oid=%p[%d] => J{", oid, size);
    for (i = 0; i < size; i++) {  
	jint u41 = get_u4 (p);
	jint u42 = get_u4 (p);
	if (i > 0)
	    fprintf (fp, ", ");
	if (u42 != 0) 
	    fprintf (fp, "0x%08x%08x", u42, u41);
	else
	    fprintf (fp, "0x%d", u41);
    }
    fprintf (fp, "}\n"); 
}

static void dump_floats (FILE* fp, jobjectID oid, jint size, char** p) {
    int i;
    fprintf (fp, "oid=%p[%d] => F{", oid, size);
    for (i = 0; i < size; i++) {  
	jint u4 = get_u4 (p);
	if (i > 0)
	    fprintf (fp, ", ");
	fprintf (fp, "%f", (float)u4);
    }
    fprintf (fp, "}\n");     
}

static void dump_doubles (FILE* fp, jobjectID oid, jint size, char** p) {
    int i;
    fprintf (fp, "oid=%p[%d] => D{", oid, size);
    for (i = 0; i < size; i++) {  
	jint u41 = get_u4 (p);
	jint u42 = get_u4 (p);
	if (i > 0)
	    fprintf (fp, ", ");
	fprintf (fp, "0x%d%d", u42, u41);
    }
    fprintf (fp, "}\n");     
}

static void dump_objects (FILE* fp, jobjectID oid, jobjectID clz, 
			  jint size, char** p, hashtab* heap_hash) {
    int i;
    cls* c = get_class (clz);
    fprintf (fp, "oid=%p[%d] => %p,%s{", oid, size, clz, c ? c->name : "<unknown>");
    for (i = 0; i < size; i++) {  
	jobjectID o = get_pointer (p);
	if (i > 0)
	    fprintf (fp, ", ");
	if (o)
	    link_object (o, oid, clz, ARRAY, -1, i, heap_hash);
	fprintf (fp, "%p", o);
    }
    fprintf (fp, "}\n");     
}

static jobjectID get_variable (FILE* fp, char** p, field* f) {
    jobjectID o = NULL;
    fprintf (fp, "%s -> ", f->field_name);
    switch (f->type) {
    case JVMPI_NORMAL_OBJECT: {
	o = get_pointer (p);
	fprintf (fp, "L%p", o);
	break;
    }
    case JVMPI_CLASS: {
	o = get_pointer (p);
	fprintf (fp, "L[]%p", o);
	break;
    }
    case JVMPI_BOOLEAN: {
	unsigned char u1;
	u1 = get_u1 (p);
	fprintf (fp, "Z%d", u1);
	break;
    }
    case JVMPI_BYTE: {
	unsigned char u1;
	u1 = get_u1 (p);
	fprintf (fp, "B%d", u1);
	break;
    }
    case JVMPI_CHAR: {
	jshort u2;
	u2 = get_u2 (p);
	fprintf (fp, "C%d", u2);
	break;
    }
    case JVMPI_SHORT: {
	jshort u2;
	u2 = get_u2 (p);
	fprintf (fp, "S%d", u2);
	break;
    }
    case JVMPI_INT: {
	jint u4;
	u4 = get_u4 (p);
	fprintf (fp, "I%d", u4);
	break;
    }
    case JVMPI_LONG: {
	jint  u41, u42;
	u41 = get_u4 (p);
	u42 = get_u4 (p);
	fprintf (fp, "J%08x%08x", u42, u41);
	break;
    }
    case JVMPI_FLOAT: {
	jfloat u4;
	u4 = get_u4 (p);
	fprintf (fp, "F%f", u4);
	break;
    }
    case JVMPI_DOUBLE: {
	jint  u41, u42;
	u41 = get_u4 (p);
	u42 = get_u4 (p);
	fprintf (fp, "D%x%x", u42, u41);
	break;
    }
    }
    fprintf (fp, "\n");
    return o;
}

void heap_dump_0 (int dump_level,
		  char *begin,
		  char *end,
		  jint num_traces,
		  JVMPI_CallTrace *traces) {
    char type;
    jobjectID oid;
    obj* o;
    size_t c = 0;
    char *p = begin;
    
    while (p < end) {	
	type = *p++;
	oid = get_pointer (&p);
	o = get_object (oid);
	if (!o)
	    get_object_alloc (oid);
	c++;
    }    
    check_objects (c);
}

void heap_dump_1 (int dump_level,
		  char *begin,
		  char *end,
		  jint num_traces,
		  JVMPI_CallTrace *traces) {
    heap_dump_2 (dump_level, begin, end, num_traces, traces);
}

void heap_dump_2 (int dump_level,
		  char *begin,
		  char *end,
		  jint num_traces,
		  JVMPI_CallTrace *traces) {

    unsigned char type, ty;
    
    jobjectID oid, clz;
    jobject jobj;
    jint u4, i;
    JNIEnv* env;
    cls* c;
    hashtab* heap_hash;
    size_t counter = 0;
    char* p = begin;    
    size_t instances = 0;
    FILE* fp = get_heap_dump_file ();
    
    fprintf (fp, "heap_dump: %d, %p, %p, %d, %p\n", 
	     dump_level, begin, end, num_traces, traces);
    fprintf (fp, "void*; %d, u4: %d, u2: %d, u1: %d\n", 
	     sizeof (void*), sizeof (jint), sizeof (unsigned short), sizeof (unsigned char));
    
    heap_hash = hash_new (20000, heap_hash_f, heap_hash_cmp);
    while (p < end) {
	/* make the ui somewhat responsible...*/
	if ((counter++ % 100) == 0 && gtk_events_pending ()) 
	    gtk_main_iteration ();
	type = *p++;
	fprintf (fp, "**********************************\n"
		 "%x: p = %p (end: %p) type: %d\n", (p - begin), p, end, type);
	fflush (fp);
	switch (type) {
	case JVMPI_GC_ROOT_UNKNOWN:
	    oid = get_pointer (&p);
	    fprintf (fp, "gc_root_unknown: oid=%p\n", oid);
	    break;
	case JVMPI_GC_ROOT_JNI_GLOBAL:
	    oid = get_pointer (&p);
	    jobj = get_pointer (&p);  
	    fprintf (fp, "gc_root_jni_global: oid=%p, object=%p\n", oid, jobj);
	    break;
	case JVMPI_GC_ROOT_JNI_LOCAL:
	    dump_triplet (fp, "gc_root_jni_local", &p);
	    break;
	case JVMPI_GC_ROOT_JAVA_FRAME:
	    dump_triplet (fp, "gc_root_java_frame", &p);
	    break;
	case JVMPI_GC_ROOT_NATIVE_STACK:
	    oid = get_pointer (&p);
	    env = get_pointer (&p);
	    fprintf (fp, "gc_root_native_stack: oid=%p, env=%p\n", oid, env);
	    break;
	case JVMPI_GC_ROOT_STICKY_CLASS:
	    oid = get_pointer (&p);
	    fprintf (fp, "gc_root_sticky_class: oid=%p\n", oid);	    
	    break;
	case JVMPI_GC_ROOT_THREAD_BLOCK:
	    oid = get_pointer (&p);
	    env = get_pointer (&p);
	    fprintf (fp, "gc_root_thread_block: oid=%p, env=%p\n", oid, env);
	    break;
	case JVMPI_GC_ROOT_MONITOR_USED:
	    oid = get_pointer (&p);
	    fprintf (fp, "gc_root_monitor_used: oid=%p\n", oid);
	    break;
	case JVMPI_GC_ROOT_THREAD_OBJ:
	    fprintf (fp, "gc_root_thread_obj is undocumented\n");
	    break;
	case JVMPI_GC_CLASS_DUMP: {
	    jobjectID super, cls_loader, signers;
	    jobjectID protection_domain, class_name, ifacep;
	    void* reserved;
	    jint instance_size;
	    jshort constant_pool_size;
	    cls* supercls;
	    cls* iface;

	    instances++;
	    clz = get_pointer (&p);
	    c = get_class (clz);
	    if (c == NULL) {
		if (clz != NULL)
		    get_class_load (clz);
		c = get_class (clz);
		if (c == NULL) {
		    fprintf (stderr, "doh, bad class in heap dump... %p\n", clz);
		    return;
		}
	    }
	    super = get_pointer (&p);
	    supercls = get_class (super);
	    cls_set_super (c, supercls);
	    cls_loader = get_pointer (&p);
	    signers = get_pointer (&p);
	    protection_domain = get_pointer (&p);
	    class_name = get_pointer (&p);
	    reserved = get_pointer (&p);
	    instance_size = get_u4 (&p);
	    
	    fprintf (fp, "class_dump: class %p (%p) => %s, %p => %s, %d\n", 
		     clz, c, c->name,  super, (supercls ? supercls->name : ""), 
		     instance_size);
	    /* read the interfaces. */
	    for (i = 0; i < c->n_interfaces; i++) {
		ifacep = get_pointer (&p);
		iface = get_class (ifacep);
		fprintf (fp, "\t%p => %s\n", ifacep, 
			 (iface ? iface->name : "?"));
	    }	    

	    constant_pool_size = get_u2 (&p);
	    fprintf (fp, "\tconstant pool size: %d\n", constant_pool_size);
	    for (i = 0; i < constant_pool_size; i++) {
		jobjectID const_pool_obj;
		jshort constant_pool_index = get_u2 (&p);
		ty = get_u1 (&p);
		const_pool_obj = get_pointer (&p);
		fprintf (fp, "\tconstant pool[%d] => %d, %p\n", i, 
			 constant_pool_index, const_pool_obj);
	    }
	    
	    fprintf (fp, "static variables\n");
	    for (i = 0; i < c->n_statics; i++) {
		jobjectID sobj = get_variable (fp, &p, &c->statics[i]);
		if (sobj) 
		    link_object (sobj, clz, clz, STATIC_VARIABLE, i, -1, heap_hash);
	    }
	    fprintf (fp, "class done\n");
	    break;
	}
	case JVMPI_GC_INSTANCE_DUMP: {
	    char* pe;
	    cls* supercls;
	    instances++;
	    oid = get_pointer (&p);
	    clz = get_pointer (&p);
	    c = get_class (clz);
	    if (c == NULL) {
		get_class_load (clz);
		c = get_class (clz);
		if (c == NULL)
		    fprintf (stderr, "class is null for %p....\n", clz);
	    }
	    u4 = get_u4 (&p);
	    pe = p + u4;
	    fprintf (fp, "instance dump (%p): %p => %s, %d\n", 
		     oid, clz, (c ? c->name : "?"), u4);
	    while (c != NULL) {
		jobjectID sobj;
		fprintf (fp, "class '%s' has %d variables...\n", 
			 c->name, c->n_instances);
		for (i = 0; i < c->n_instances; i++) {
		    sobj = get_variable (fp, &p, &c->instances[i]);
		    if (sobj)
			link_object (sobj, oid, c->class_id, VARIABLE, i, -1, heap_hash);
		}
		supercls = cls_get_super (c);
		if (supercls == NULL) 
		    get_super_class (c);
		c = cls_get_super (c);
	    }
	    if (p != pe) {
		fprintf (stderr, "doh, failed to read instance variables size: %d.....\n", u4);
		p = pe;
	    }
	    fprintf (fp, "instance done...\n");
	    break;
	}
	case JVMPI_GC_OBJ_ARRAY_DUMP:
	    instances++;
	    oid = get_pointer (&p);
	    u4 = get_u4 (&p);
	    clz = get_pointer (&p);
	    dump_objects (fp, oid, clz, u4, &p, heap_hash);
	    break;
	case JVMPI_GC_PRIM_ARRAY_DUMP:
	    instances++;
	    oid = get_pointer (&p);
	    u4 = get_u4 (&p);
	    ty = get_u1 (&p);
	    switch (ty) {
	    case JVMPI_BOOLEAN:
		dump_booleans (fp, oid, u4, &p);
		break;
	    case JVMPI_BYTE:
		dump_bytes (fp, oid, u4, &p);
		break;
	    case JVMPI_CHAR:
		dump_chars (fp, oid, u4, &p);
		break;
	    case JVMPI_SHORT:
		dump_shorts (fp, oid, u4, &p);
		break;
	    case JVMPI_INT:
		dump_ints (fp, oid, u4, &p);
		break;
	    case JVMPI_LONG:
		dump_longs (fp, oid, u4, &p);
		break;
	    case JVMPI_FLOAT:
		dump_floats (fp, oid, u4, &p);
		break;
	    case JVMPI_DOUBLE:
		dump_doubles (fp, oid, u4, &p);
		break;
	    }
	    break;
	default:
	    fprintf (fp, "unknown type..."); 
	    exit (-1);
	}
	fflush (fp);
    }

    fprintf (stderr, "heap_hash has: %d entries...\n", hash_cardinal (heap_hash));
    clear_and_set_last_heap (heap_hash);
    check_objects (instances);
}

/** Handle object dumps, currently only used to get super classes.. 
 */
void object_dump (jint data_len, char *data) {
    unsigned char type;
    char* p = data;
    type = *p++;
    if (type == JVMPI_GC_CLASS_DUMP) {
	cls* c = NULL, * s = NULL;
	jobjectID oc = get_pointer (&p);
	jobjectID os = get_pointer (&p);
	c = get_class (oc);
	if (os != NULL) {
	    s = get_class (os);
	    if (s == NULL) {
		/*fprintf (stderr, "unknown class (%p).. requesting..\n", os);*/
		get_class_load (os);
		s = get_class (os);
	    }
	}
	if (c == NULL) {
	    fprintf (stderr, "doh: %p => no class found...\n", oc);
	} else {
	    cls_set_super (c, s);
	}
    }    
}


/** Get the owners of the given object at the last heap dump.
 * @param obj the object to get the owners for.
 * @return an linked list of object_link containing the owners of obj.
 */
object_link* get_owners (jobjectID obj) {
    static object_link ol;
    object_link* ret;
    if (last_heap == NULL)
	return NULL;
    ol.obj = obj;
    ret = hash_search (&ol, last_heap);
    return ret;
}

int is_get_owners_possible () {
    return last_heap != NULL;
}
